@@ -1,17 +1,14 @@
 #################################################################
 # GDAL wrapper for convenient raster data handling and processing
-# John Truckenbrodt 2015-2020
+# John Truckenbrodt 2015-2021
 #################################################################
-
-
-# todo: function to write data with the same metadata as a given file
-
 from __future__ import division
 import os
 import re
 import platform
 import warnings
 import tempfile
+from datetime import datetime
 from math import sqrt, floor, ceil
 from time import gmtime, strftime
 import numpy as np
@@ -66,9 +63,11 @@ class Raster(object):
     list_separate: bool
         treat a list of files as separate layers or otherwise as a single layer? The former is intended for single
         layers of a stack, the latter for tiles of a mosaic.
+    timestamps: list[str] or function or None
+        the time information for each layer or a function converting band names to a :obj:`datetime.datetime` object
     """
     
-    def __init__(self, filename, list_separate=True):
+    def __init__(self, filename, list_separate=True, timestamps=None):
         if isinstance(filename, gdal.Dataset):
             self.raster = filename
             self.filename = self.files[0] if self.files is not None else None
@@ -93,7 +92,8 @@ class Raster(object):
         self.__data = [None] * self.bands
         
         if self.format == 'ENVI':
-            with HDRobject(self.filename + '.hdr') as hdr:
+            hdrfile = os.path.splitext(re.sub('/vsi(?:zip|tar)/', '', self.filename))[0] + '.hdr'
+            with HDRobject(hdrfile) as hdr:
                 if hasattr(hdr, 'band_names'):
                     self.bandnames = hdr.band_names
                 else:
@@ -106,6 +106,15 @@ class Raster(object):
                 self.bandnames = ['mosaic']
         else:
             self.bandnames = ['band{}'.format(x) for x in range(1, self.bands + 1)]
+        
+        if isinstance(timestamps, list):
+            if len(timestamps) != len(self.bandnames):
+                raise RuntimeError('the number of time stamps is different to the number of bands')
+            self.timestamps = timestamps
+        elif callable(timestamps):
+            self.timestamps = [timestamps(x) for x in self.bandnames]
+        else:
+            self.timestamps = None
     
     def __enter__(self):
         return self
@@ -120,13 +129,21 @@ class Raster(object):
         vals['proj4'] = self.proj4
         vals['epsg'] = self.epsg
         vals['filename'] = self.filename if self.filename is not None else 'memory'
+        if self.timestamps is not None:
+            t0 = min(self.timestamps)
+            t1 = max(self.timestamps)
+            vals['time'] = 'time range : {0} .. {1}\n'.format(t0, t1)
+        else:
+            vals['time'] = ''
         
         info = 'class      : spatialist Raster object\n' \
                'dimensions : {rows}, {cols}, {bands} (rows, cols, bands)\n' \
                'resolution : {xres}, {yres} (x, y)\n' \
+               '{time}' \
                'extent     : {xmin}, {xmax}, {ymin}, {ymax} (xmin, xmax, ymin, ymax)\n' \
                'coord. ref.: {proj4} (EPSG:{epsg})\n' \
                'data source: {filename}'.format(**vals)
+        
         return info
     
     def __getitem__(self, index):
@@ -186,6 +203,9 @@ class Raster(object):
         coord. ref.: +proj=utm +zone=29 +datum=WGS84 +units=m +no_defs
         data source: /tmp/tmps4rc9o09.vrt
         """
+        if not isinstance(index, (slice, tuple, Vector)):
+            raise TypeError('index must be of type tuple, slice or spatialist.Vector')
+        
         # subsetting via Vector object
         if isinstance(index, Vector):
             geomtypes = list(set(index.geomTypes))
@@ -204,7 +224,7 @@ class Raster(object):
             # subset raster object with slices
             with self[sl] as sub:
                 # mask subsetted raster object with vector geometries
-                masked = sub.__maskbyvector(inter)
+                masked = sub.__maskbyvector(inter, nodata=self.nodata)
             inter = None
             return masked
         #####################################################################################################
@@ -216,7 +236,8 @@ class Raster(object):
                     'mismatch of index length ({0}) and raster dimensions ({1})'.format(len(index), ras_dim))
             for i in [0, 1]:
                 if hasattr(index[i], 'step') and index[i].step is not None:
-                    raise IndexError('step slicing of {} is not allowed'.format(['rows', 'bands'][i]))
+                    raise IndexError('step slicing of {} is not allowed'.format(['rows', 'cols'][i]))
+            index = list(index)
         
         # treat float indices as map coordinates and convert them to image coordinates
         yi = index[0]
@@ -245,6 +266,44 @@ class Raster(object):
             if not isinstance(subset[key], list):
                 subset[key] = [subset[key]]
         if len(index) > 2:
+            bi = index[2]
+            if isinstance(bi, str):
+                bi = self.bandnames.index(bi)
+            elif isinstance(bi, datetime):
+                bi = self.timestamps.index(bi)
+            elif isinstance(bi, int):
+                pass
+            elif isinstance(bi, slice):
+                if isinstance(bi.start, int):
+                    start = bi.start
+                elif isinstance(bi.start, str):
+                    start = self.bandnames.index(bi.start)
+                elif isinstance(bi.start, datetime):
+                    larger = [x for x in self.timestamps if x > bi.start]
+                    tdiff = [x - bi.start for x in larger]
+                    closest = larger[tdiff.index(min(tdiff))]
+                    start = self.timestamps.index(closest)
+                elif bi.start is None:
+                    start = None
+                else:
+                    raise TypeError('band indices must be either int or str')
+                if isinstance(bi.stop, int):
+                    stop = bi.stop
+                elif isinstance(bi.stop, str):
+                    stop = self.bandnames.index(bi.stop)
+                elif isinstance(bi.stop, datetime):
+                    smaller = [x for x in self.timestamps if x < bi.stop]
+                    tdiff = [bi.start - x for x in smaller]
+                    closest = smaller[tdiff.index(min(tdiff))]
+                    stop = self.timestamps.index(closest)
+                elif bi.stop is None:
+                    stop = None
+                else:
+                    raise TypeError('band indices must be either int or str')
+                bi = slice(start, stop)
+            else:
+                raise TypeError('band indices must be either int or str')
+            index[2] = bi
             subset['bands'] = list(range(0, self.bands))[index[2]]
             if not isinstance(subset['bands'], list):
                 subset['bands'] = [subset['bands']]
@@ -275,14 +334,22 @@ class Raster(object):
         # outname = os.path.join('/vsimem/', os.path.basename(tempfile.mktemp()))
         outname = tempfile.NamedTemporaryFile(suffix='.vrt').name
         out_ds = gdalbuildvrt(src=self.filename, dst=outname, options=opts, void=False)
-        out = Raster(out_ds)
+        
+        timestamps = self.timestamps
         if len(index) > 2:
             bandnames = self.bandnames[index[2]]
+            if timestamps is not None:
+                timestamps = timestamps[index[2]]
         else:
             bandnames = self.bandnames
         if not isinstance(bandnames, list):
             bandnames = [bandnames]
+        if timestamps is not None and not isinstance(timestamps, list):
+            timestamps = [timestamps]
+        
+        out = Raster(out_ds)
         out.bandnames = bandnames
+        out.timestamps = timestamps
         return out
     
     def __extent2slice(self, extent):
@@ -322,7 +389,8 @@ class Raster(object):
                                    self.cols, self.rows, self.bands, Dtype(self.dtype).gdalint)
         driver = None
         outDataset.SetMetadata(self.raster.GetMetadata())
-        outDataset.SetGeoTransform([self.geo[x] for x in ['xmin', 'xres', 'rotation_x', 'ymax', 'rotation_y', 'yres']])
+        keys = ['xmin', 'xres', 'rotation_x', 'ymax', 'rotation_y', 'yres']
+        outDataset.SetGeoTransform([self.geo[x] for x in keys])
         if self.projection is not None:
             outDataset.SetProjection(self.projection)
         for i in range(1, self.bands + 1):
@@ -332,6 +400,14 @@ class Raster(object):
             mat = mat * mask
             outband.WriteArray(mat)
             del mat
+            
+            inband = self.raster.GetRasterBand(i)
+            cmap = inband.GetRasterColorTable()
+            if cmap is not None:
+                outband.SetRasterColorTable(cmap)
+                cmap = None
+            inband = None
+            
             outband.FlushCache()
             outband = None
         if outname is not None:
@@ -345,6 +421,7 @@ class Raster(object):
         else:
             out = Raster(outDataset)
             out.bandnames = self.bandnames
+            out.timestamps = self.timestamps
             return out
     
     def __prependVSIdirective(self, filename):
@@ -353,7 +430,7 @@ class Raster(object):
         
         Parameters
         ----------
-        filename: str
+        filename: str or list
             the file name, e.g. archive.tar.gz/filename
 
         Returns
@@ -361,9 +438,9 @@ class Raster(object):
 
         """
         if isinstance(filename, str):
-            if re.search(r'\.zip', filename):
+            if re.search(r'\.zip', filename) and not re.search('/vsizip/', filename):
                 filename = '/vsizip/' + filename
-            if re.search(r'\.tar', filename):
+            if re.search(r'\.tar', filename) and not re.search('/vsitar/', filename):
                 filename = '/vsitar/' + filename
         elif isinstance(filename, list):
             filename = [self.__prependVSIdirective(x) for x in filename]
@@ -986,8 +1063,8 @@ class Raster(object):
         """
         return osr.SpatialReference(wkt=self.projection)
     
-    def write(self, outname, dtype='default', format='ENVI', nodata='default', compress_tif=False, overwrite=False,
-              cmap=None, update=False, xoff=0, yoff=0, array=None):
+    def write(self, outname, dtype='default', format='ENVI', nodata='default', overwrite=False,
+              cmap=None, update=False, xoff=0, yoff=0, array=None, options=None, overviews=None):
         """
         write the raster object to a file.
 
@@ -1017,6 +1094,10 @@ class Raster(object):
             the y/row offset
         array: numpy.ndarray
             write different data than that associated with the Raster object
+        options: list or None
+            as list of options for creating the output dataset; see :osgeo:meth:`gdal.Driver.Create`
+        overviews: list or None
+            a list of integer overview levels to be created; see :osgeo:meth:`gdal.Dataset.BuildOverviews`
 
         Returns
         -------
@@ -1033,9 +1114,8 @@ class Raster(object):
             
             nodata = self.nodata if nodata == 'default' else nodata
             
-            options = []
-            if format == 'GTiff' and compress_tif:
-                options += ['COMPRESS=DEFLATE', 'PREDICTOR=2']
+            if options is None:
+                options = []
             
             driver = gdal.GetDriverByName(format)
             outDataset = driver.Create(outname, self.cols, self.rows, self.bands, dtype, options)
@@ -1051,8 +1131,19 @@ class Raster(object):
             outband = outDataset.GetRasterBand(i)
             
             if not update_existing:
-                if cmap is not None:
-                    outband.SetRasterColorTable(cmap)
+                if cmap is None:
+                    inband = self.raster.GetRasterBand(i)
+                    colmap = inband.GetRasterColorTable()
+                    inband = None
+                else:
+                    colmap = cmap
+                try:
+                    outband.SetRasterColorTable(colmap)
+                except RuntimeError as e:
+                    if cmap is not None:
+                        warnings.warn(str(e))
+                    else:
+                        pass
                 if nodata is not None:
                     outband.SetNoDataValue(nodata)
             if array is None:
@@ -1076,6 +1167,8 @@ class Raster(object):
             outband = None
         if format == 'GTiff':
             outDataset.SetMetadataItem('TIFFTAG_DATETIME', strftime('%Y:%m:%d %H:%M:%S', gmtime()))
+        if overviews:
+            outDataset.BuildOverviews('NEAREST', overviews)
         outDataset = None
         if format == 'ENVI':
             hdrfile = os.path.splitext(outname)[0] + '.hdr'
